import bisect
import io
import multiprocessing.pool
import os
import signal
from typing import List, Optional

try:
    import xz
except ImportError:
    xz = None  # type: ignore

try:
    import indexed_zstd
except ImportError:
    indexed_zstd = None  # type: ignore

from .utils import LRUCache, Prefetcher, overrides

_parallelXzReaderFile = None
_parallelZstdReaderFile = None


class BlockParallelReader(io.BufferedIOBase):
    """Uses a process pool to prefetch and cache decompressed blocks"""

    # Using a thread pool, slowed down ParallelXZReader by 50%

    def __init__(
        self,
        filename: str,
        fileobj,
        blockBoundaries: List[int],
        parallelization: Optional[int],
        initWorker=None,
        initArgs=(),
    ):
        """
        blockBoundaries:
            All but the last block boundary are also the uncompressed offsets at which new blocks begin
            The last block boundary is the first past-the-file-end offset (i.e. requal to the file size)
            and can be used to deduce the size of the last block.
        """
        if not parallelization:
            parallelization = os.cpu_count()
            assert parallelization is not None, "Cannot automatically determine CPU count!"
        # keep one core for on-demand decompression
        self.parallelization: int = parallelization - 1
        assert self.parallelization >= 1, "If you do not need to parallelize, then do not use this class!"

        self.filename = filename
        self.fileobj = fileobj
        self.blockBoundaries: List[int] = blockBoundaries
        self.initWorker = BlockParallelReader._init_worker if initWorker is None else initWorker
        self.initArgs = initArgs

        self._pool = None
        self._offset = 0
        self._blockCache: LRUCache = LRUCache(2 * parallelization)
        self._prefetcher = Prefetcher(4)
        self._lastUsedBlock: int = 0

        self.requestCount: int = 0
        self.cacheHitCount: int = 0
        self.cacheMissCount: int = 0
        self.cachePrefetchCount: int = 0

        assert self.fileobj.seekable()
        assert self.fileobj.readable()

    def join_threads(self):
        self._pool.close()
        self._pool.join()
        self._pool = None

    def _get_pool(self):
        if not self._pool:
            self._pool = multiprocessing.pool.Pool(self.parallelization, self.initWorker, self.initArgs)
        return self._pool

    @staticmethod
    def _init_worker():
        """
        Ignore the interrupt signal inside the child worker processes to avoid Python backtraces for each of them.
        Aborting with Ctrl+C will still work as the main process still accepts the signal.
        """
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    @staticmethod
    def _find_block(blockBoundaries: List[int], offset: int) -> Optional[int]:
        """
        Returns the blockNumber such that the requested offset is inside the block starting at
        self.blockBoundaries[blockNumber].
        """
        blockNumber = bisect.bisect_right(blockBoundaries, offset) - 1
        # If the found block is the last element, then the offset is actually after the EOF!
        return blockNumber if blockNumber >= 0 and blockNumber + 1 < len(blockBoundaries) else None

    @staticmethod
    def _block_size(blockBoundaries: List[int], blockNumber) -> int:
        if blockNumber + 1 >= len(blockBoundaries) or blockNumber < 0:
            return 0
        return blockBoundaries[blockNumber + 1] - blockBoundaries[blockNumber]

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    @overrides(io.BufferedIOBase)
    def close(self) -> None:
        self.fileobj.close()
        if self._pool:
            self._pool.close()

    @overrides(io.BufferedIOBase)
    def fileno(self) -> int:
        # This is a virtual Python level file object and therefore does not have a valid OS file descriptor!
        raise io.UnsupportedOperation

    @overrides(io.BufferedIOBase)
    def seekable(self) -> bool:
        return True

    @overrides(io.BufferedIOBase)
    def readable(self) -> bool:
        return True

    @overrides(io.BufferedIOBase)
    def writable(self) -> bool:
        return False

    @overrides(io.BufferedIOBase)
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if whence == io.SEEK_CUR:
            self._offset += offset
        elif whence == io.SEEK_END:
            self._offset = self.blockBoundaries[-1] + offset
        elif whence == io.SEEK_SET:
            self._offset = offset

        if self._offset < 0:
            raise ValueError("Trying to seek before the start of the file!")

        return self._offset

    @overrides(io.BufferedIOBase)
    def tell(self) -> int:
        return self._offset

    def _read(self, size: int, decodeBlock) -> bytes:
        blocks = []
        blockNumber = self._find_block(self.blockBoundaries, self._offset)
        if blockNumber is None:
            return b""

        self.requestCount += 1
        firstBlockOffset = self._offset - self.blockBoundaries[blockNumber]
        blockSize = self._block_size(self.blockBoundaries, blockNumber)
        availableSize = blockSize - firstBlockOffset

        # Shortcut to improve performance for many small reads inside the same block
        # Alternatively, it might work to wrap this in an io.BufferedRandom to yield larger reads.
        # But, I think this is a small enough workaround for such a large impact.
        if blockNumber in self._blockCache and size <= availableSize and self._lastUsedBlock == blockNumber:
            self.cacheHitCount += 1
            result = self._blockCache[blockNumber].get()[firstBlockOffset : firstBlockOffset + size]
            self._offset += len(result)
            return result

        pendingBlocks: int = sum(0 if block.ready() else 1 for block in self._blockCache.values())
        pool = self._get_pool()

        # TODO do not fetch blocks which do not fit into memory and instead seek directly to them!
        #      Note that this only makes sense if less than the block size has been requested else we would run out
        #      of memory in botch events! So it makes only sense for the first and last requested block.
        while True:
            # Fetch Block
            self._lastUsedBlock = blockNumber
            self._prefetcher.fetch(blockNumber)
            if blockNumber in self._blockCache:
                fetchedBlock = self._blockCache[blockNumber]
            else:
                self.cachePrefetchCount += 1
                fetchedBlock = pool.apply_async(
                    decodeBlock,
                    (
                        self.filename,
                        self.blockBoundaries[blockNumber],
                        self._block_size(self.blockBoundaries, blockNumber),
                    ),
                )
                self._blockCache[blockNumber] = fetchedBlock
                pendingBlocks += 1

            blocks.append(fetchedBlock)
            self._offset += min(availableSize, size)
            if size <= availableSize:
                break

            # Get metadata for next block
            blockNumber += 1
            if blockNumber + 1 >= len(self.blockBoundaries):
                break

            # Only decrement size when we are actually entering the next iteration
            size -= availableSize

            # Get block data for next iteration
            blockSize = self._block_size(self.blockBoundaries, blockNumber)
            offsetInBlock = self._offset - self.blockBoundaries[blockNumber]
            availableSize = blockSize - offsetInBlock

        # Prefetch blocks
        toPrefetch = self._prefetcher.prefetch(self.parallelization)
        for blockToPrefetch in toPrefetch:
            blockSize = self._block_size(self.blockBoundaries, blockToPrefetch)
            if blockSize > 0 and blockToPrefetch not in self._blockCache and pendingBlocks < self.parallelization:
                self.cachePrefetchCount += 1
                fetchedBlock = pool.apply_async(
                    decodeBlock,
                    (self.filename, self.blockBoundaries[blockToPrefetch], blockSize),
                )
                self._blockCache[blockToPrefetch] = fetchedBlock
                pendingBlocks += 1

        # Concatenate data from blocks as necessary
        result = b""
        while blocks:
            block = blocks.pop(0)
            if not block.ready():
                self.cacheMissCount += 1
            # Note that it is perfectly safe to call AsyncResult.get multiple times!
            toAppend = block.get()
            if firstBlockOffset > 0:
                toAppend = toAppend[firstBlockOffset:]
            if not blocks:
                toAppend = toAppend[:size]
            firstBlockOffset = 0

            result += toAppend

        # TODO fall back to reading directly from fileobj if prefetch suggests nothing at all to improve latency!
        # self.fileobj.seek(self._offset)
        # result = self.fileobj.read(size)

        return result


class ParallelXZReader(BlockParallelReader):
    def __init__(self, filename: str, parallelization: Optional[int] = None):
        fileObject = xz.open(filename, 'rb')

        # The block boundaries in the decompressed stream are computes as such: [
        #     stream_pos + block_boundary
        #     for stream_pos, stream in self._fileobjs.items()
        #     for block_boundary in stream.block_boundaries
        # ]
        # This also gives us hints how to manually compute the compressed block offsets.
        blockBoundaries: List[int] = fileObject.block_boundaries.copy()
        blockBoundaries.append(len(fileObject))

        # Compute compressed block offsets because python-xz has no way to query that information:
        # https://github.com/Rogdham/python-xz/blob/89af850a59aaf83920a0eb7c314d9f2ed71979fa/src/xz/stream.py#L43-L46
        # These offsets are only approximate because of the missing interface and because I do not want to parse
        # the format myself.
        self.approximateCompressedBlockBoundaries: List[int] = []
        if hasattr(fileObject, '_fileobjs'):
            streamOffset = 0
            # xz.open returns XZFile defined in xz/file.py as subclass of IOCombiner[XZStream].
            # fileObject._fileobjs: FloorDict[T] where T=XZStream is part of IOCombiner defined in xz/io.py.
            # It maps decompressed offsets to XZStream objects.
            for stream in fileObject._fileobjs.values():
                # https://tukaani.org/xz/xz-file-format.txt
                # The stream header and footer are both 12 B. Streams are padded to the next offset divisible by 4.
                blockOffset = 12
                for block in stream._fileobjs.values():
                    self.approximateCompressedBlockBoundaries.append(streamOffset + blockOffset)
                    # This is very likely not accurate. In tests, the sum of block sizes does not add up
                    # to the stream size! This is probably because of the index comes after the last block,
                    # and the padding after the stream footer.
                    # E.g. in tests: Sum of block sizes: 49654928 stream size: 49655064
                    # The difference is 136, subtracting the strea header and footer: 112.
                    # There are 14 blocks per stream and 112 / 14 = 8. This may be a coincidence because:
                    #  - The index has CRC32 and other extra bytes and might use variable length integers.
                    #  - Python-xz seems to wrap each XZ block in an XZ stream header and footer to decode
                    #    it as XZ with liblzma. I'm not sure whether this overhead is returned when calling
                    #    len on the file object.
                    blockOffset += len(block.fileobj)
                streamOffset += len(stream.fileobj)
            # This is necessary because we also appended the decompressed file size to blockBoundaries above!
            self.approximateCompressedBlockBoundaries.append(streamOffset)
            # Clear list if consistency check fails. This is currently only used for the progress indicator anyway.
            if (
                len(blockBoundaries) != len(self.approximateCompressedBlockBoundaries)
                or streamOffset != os.stat(filename).st_size
            ):
                self.approximateCompressedBlockBoundaries = []

        super().__init__(
            filename, fileObject, blockBoundaries, parallelization, ParallelXZReader._init_worker2, (filename,)
        )
        self._open_files()

    def _open_files(self):
        # Opening the pool and and the files on each worker at this point might be a point to discuss
        # but it leads to uniform latencies for the subsequent read calls.
        pool = self._get_pool()
        # will trigger worker initialization, i.e., _try_open_global_file
        results = [pool.apply_async(os.getpid) for _ in range(self.parallelization * 4)]
        for result in results:
            result.get()

    @staticmethod
    def _init_worker2(filename):
        BlockParallelReader._init_worker()
        ParallelXZReader._try_open_global_file(filename)

    @staticmethod
    def _try_open_global_file(filename):
        # This is not thread-safe! But it will be executed in a process pool, in which each worker has its own
        # global variable set. Using a global variable for this is safe because we know that there is one process pool
        # per BlockParallelReader, meaning the filename is a constant for each worker.
        # pylint: disable=global-statement
        global _parallelXzReaderFile
        if _parallelXzReaderFile is None:
            _parallelXzReaderFile = xz.open(filename, 'rb')

    @staticmethod
    def _decode_block(filename, offset, size):
        ParallelXZReader._try_open_global_file(filename)
        _parallelXzReaderFile.seek(offset)
        return _parallelXzReaderFile.read(size)

    @overrides(io.BufferedIOBase)
    def read(self, size: int = -1) -> bytes:
        return super()._read(size, ParallelXZReader._decode_block)

    def find_block(self, offset) -> Optional[int]:
        """
        Returns the block number corresponding to the specified offset in the decompressed stream.
        """
        return self._find_block(self.blockBoundaries, self._offset)


# This one is actually mostly slower than serial decoding. Even the command line tool zstd is often slower with
# -T 16 compared to -T 1 for frame sizes 1 MiB up to 512 MiB. It might simply be that zstd decompression is too
# fast to parallelize with this approach and instead leads to more cache problems or even the file reading might
# be the limiting factor. I might have to give this up for now.
class ParallelZstdReader(BlockParallelReader):
    def __init__(self, filename: str, parallelization: Optional[int] = None):
        fileObject = indexed_zstd.IndexedZstdFile(filename)
        blockBoundaries = list(fileObject.block_offsets().values())
        super().__init__(filename, fileObject, blockBoundaries, parallelization)

    @staticmethod
    def _decode_block(filename, offset, size):
        # This is not thread-safe! But it will be executed in a process pool, in which each worker has its own
        # global variable set. Using a global variable for this is safe because we know that there is one process pool
        # per BlockParallelReader, meaning the filename is a constant for each worker.
        # pylint: disable=global-statement
        global _parallelZstdReaderFile
        if _parallelZstdReaderFile is None:
            _parallelZstdReaderFile = indexed_zstd.IndexedZstdFile(filename)

        _parallelZstdReaderFile.seek(offset)
        return _parallelZstdReaderFile.read(size)

    @overrides(io.BufferedIOBase)
    def read(self, size: int = -1) -> bytes:
        return super()._read(size, ParallelZstdReader._decode_block)



import os
import shutil
import sqlite3
import sys
import tempfile
import urllib.parse
from typing import IO, Any, Dict, Iterable, List, Optional, Union

from ratarmountcore.formats import FileFormatID, replace_format_check
from ratarmountcore.mountsource import FileInfo, MountSource, create_root_file_info
from ratarmountcore.SQLiteBlobFile import SQLiteBlobFile
from ratarmountcore.StenciledFile import LambdaReaderFile
from ratarmountcore.utils import overrides

try:
    import rapidgzip
except ImportError:
    rapidgzip = None

# Password protection requires the paid SEE version of SQLite or sqlcipher.
#  - https://www.sqlite.org/see/doc/trunk/www/readme.wiki
#  - https://github.com/sqlcipher/sqlcipher
#
# Overview of Python bindings:
#  - https://github.com/rigglemania/pysqlcipher3
#    Last update: 2 years ago
#    > This project is no longer being actively maintained. Security vulnerabilities may exist in this code.
#    > Use at your own risk.
#    > This library is a fork of pysqlcipher targeted for use with Python 3, although support for Python 2
#    > is still maintained
#    - https://pypi.org/project/pysqlcipher3
#
#  - https://github.com/coleifer/pysqlite3
#    Last commit: 2 months ago
#    > This library takes the SQLite module from Python 3 and packages it as a separately-installable module.
#    > This may be useful for creating SQLite modules capable of working with other versions of SQLite
#    > (via the amalgamation option).
#
#  - https://github.com/coleifer/sqlcipher3
#    Last commit: 8 months ago
#    > This library takes pysqlite3 and makes some small modifications so it is suitable for use
#    > with sqlcipher (sqlite with encryption).
#    - https://pypi.org/project/sqlcipher3/
#    - https://pypi.org/project/sqlcipher3-binary/
#      - Only has 8 wheels, namely only manylinux x86_64 for each CPython version 3.6 to 3.13.
#        No support for Windows, MacOS, Power, ARM, ...
#
#    - https://github.com/laggykiller/sqlcipher3
#      Fork. Last commit: 2 months ago, 23 stars
#      > This is a fork of sqlcipher3 which adds github action for creating wheels for Windows, MacOS and Linux.
#      > The unofficial wheels from this fork are uploaded to sqlcipher3-wheels on pypi, while the official
#      > wheels are sqlcipher3-binary.
#      See issue on fork base "PR for building wheels and other improvements"
#      - https://github.com/coleifer/sqlcipher3/issues/24
#      - https://pypi.org/project/sqlcipher3-wheels/
#
# The SQLAR (SQLite Archiver) format is fortunately pretty simple. It's very similar to SQLiteIndex used for
# ratarmount but it also stores the file contents and has some fewer columns for metadata.
#
# https://www.sqlite.org/sqlar.html
# https://sqlite.org/sqlar/doc/trunk/README.md
#
# CREATE TABLE sqlar(
#     name TEXT PRIMARY KEY,  -- name of the file
#     mode INT,               -- access permissions
#     mtime INT,              -- last modification time
#     sz INT,                 -- original file size
#     data BLOB               -- compressed content
# );
#
# > Both directories and empty files have sqlar.sz==0. Directories can be distinguished from empty files
# > because directories have sqlar.data IS NULL. The file is compressed if length(sqlar.blob)<sqlar.sz
# > and is stored as plaintext if length(sqlar.blob)==sqlar.sz.
# >
# > Symbolic links have sqlar.sz set to -1, and the link target stored as a text value in the sqlar.data field.
# >
# > SQLAR uses the "zlib format" for compression.
#
# Based on the code, "name" is assumed to not start with a leading slash.
#
# Unfortunately, it has only a single "name" column instead of splitting the file path into folder and name.
# This requires linear scans of the database in order to list all files of a folder. I find that unfortunate.
# This was one of the things I
#
# One might think that there should be no file versions because the file name(path) is the primary key,
# but using denormal paths such as a, /a, //a, ./a, and so on, one could hack file versions into this format.
# Then again, the original tool woult not support looking this up because of its globbing rule, so maybe
# let's simply ignore that this could happen. The format is not very s
#
# You can even download the SQLite source code as an sqlar archive:
#  - https://www.sqlite.org/src/dir?ci=trunk
#  - https://sqlite.org/src/sqlar/sqlite.sqlar?r=release

try:
    # This package is required by many other packages, so it is almost safe to require it:
    # Required-by: asyncssh, azure-identity, azure-storage-blob, fido2, gajim, msal, oci, omemo-dr, pyOpenSSL, pyspnego,
    #              smbprotocol, types-paramiko, types-pyOpenSSL, types-redis
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
    from sqlcipher3 import dbapi2 as sqlcipher3  # type:ignore

    # No way to detect encrypted SQLAR.
    replace_format_check(FileFormatID.SQLAR, lambda x: True)
except ImportError:
    # The cryptography imports can fail pretty badly and it does not seem to be catchable :(
    #
    #    core/tests/test_formats.py:13: in <module>
    #        import ratarmountcore.mountsource.archives  # noqa: E402, F401
    #        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    #    core/ratarmountcore/mountsource/archives.py:14: in <module>
    #        from .formats.sqlar import SQLARMountSource
    #    core/ratarmountcore/mountsource/formats/sqlar.py:99: in <module>
    #        from cryptography.hazmat.primitives import hashes
    #    /usr/lib/python3/dist-packages/cryptography/hazmat/primitives/hashes.py:10: in <module>
    #        from cryptography.hazmat.bindings._rust import openssl as rust_openssl
    #    E   pyo3_runtime.PanicException: Python API call failed
    #    ------------------------- Captured stderr -------------------------
    #    ModuleNotFoundError: No module named '_cffi_backend'
    #    thread '<unnamed>' panicked at /usr/share/cargo/registry/pyo3-0.20.2/src/err/mod.rs:788:5:
    #    Python API call failed
    #    note: run with `RUST_BACKTRACE=1` environment variable to display a backtrace
    #
    # In my case, pytest pointed to the Python 3.11 version even though my system-default is 3.12.
    # Using python3 -m pytest instead of pytest fixed the error. (Re)installing cffi and cryptography in 3.11
    # might also have worked.
    sqlcipher3 = None  # type:ignore
    PBKDF2HMAC = None  # type:ignore
    default_backend = None  # type:ignore


class FixedSqliteBlob(LambdaReaderFile):
    """
    Rapidgzip expects a seek method, but it does not exist for sqlite3.Blob :'(
    This wrapper adds a seekable method among many others.
    https://docs.python.org/3/library/sqlite3.html#sqlite3.Blob
    """

    def __init__(self, blob):
        self.blob = blob

        def pread(offset: int, size: int) -> bytes:
            self.blob.seek(offset)
            return self.blob.read(size)

        super().__init__(pread, len(self.blob))


class SQLARMountSource(MountSource):
    _SQLITE_FILEINFO_COLUMNS = "rowid, mode, mtime, sz, CASE WHEN sz=-1 THEN data ELSE '' END"
    _MAGIC_BYTES = b'SQLite format 3\x00'

    def __init__(self, fileOrPath: Union[str, IO[bytes]], passwords: Optional[List[bytes]] = None, **options) -> None:
        self._temporaryFilePath: Optional[Any] = None

        magicBytes = b""
        if isinstance(fileOrPath, str):
            with open(fileOrPath, 'rb') as file:
                magicBytes = SQLARMountSource._quick_check_file(file, fileOrPath, passwords)
            self.indexFilePath = fileOrPath
        else:
            magicBytes = SQLARMountSource._quick_check_file(fileOrPath, "File object", passwords)

            # Copy to a temporary file because sqlite cannot work with Python file objects. This can be wasteful!
            fileOrPath.seek(0)
            self._temporaryFile = tempfile.NamedTemporaryFile(suffix=".ratarmount.sqlar", delete=True)
            shutil.copyfileobj(fileOrPath, self._temporaryFile.file)  # type: ignore
            self._temporaryFile.file.flush()

            self.indexFilePath = self._temporaryFile.name

        uriPath = f"file:{urllib.parse.quote(self.indexFilePath)}?mode=ro"
        connection: Optional[Any] = None
        try:
            # check_same_thread=False can be used because it is read-only and allows to enable FUSE multithreading.
            connection = sqlite3.connect(uriPath, uri=True, check_same_thread=False)
            SQLARMountSource._check_database(connection)
        except sqlite3.DatabaseError:
            if connection:
                connection.close()
                connection = None
            if passwords and sqlcipher3 is not None and PBKDF2HMAC is not None and len(magicBytes) >= 16:
                connection = self._find_password(uriPath, passwords, salt=magicBytes[:16])
            if not connection:
                raise
        assert connection
        self.connection = connection

        self.options = options

        # Check for denormal names and store a normalized list if necessary.
        self._files: dict[str, int] = {}
        if any(
            os.path.normpath(name.lstrip('/')) != name for name, in self.connection.execute("SELECT name FROM sqlar;")
        ):
            self._files = {
                os.path.normpath(name.lstrip('/')): rowid
                for name, rowid in self.connection.execute("SELECT name,rowid FROM sqlar ORDER BY rowid;")
            }

    @staticmethod
    def _quick_check_file(fileObject: IO[bytes], name: str, passwords: Optional[List[bytes]]) -> bytes:
        try:
            magicBytes = fileObject.read(len(SQLARMountSource._MAGIC_BYTES))
        finally:
            fileObject.seek(0)

        if magicBytes == SQLARMountSource._MAGIC_BYTES or (sqlcipher3 is not None and passwords):
            return magicBytes
        message = name + " is not an SQLAR file."
        if sqlcipher3 is None and passwords:
            message += " Passwords have been ignored because sqlcipher3 and cryptography are not installed."
            message += " Please install with 'pip install sqlcipher3-binary cryptography' or 'pip install sqlcipher3-wheels cryptography'."
        if sqlcipher3 is not None and not passwords:
            message += " Please specify passwords if it is an encrypted SQLAR."
        raise ValueError(message)

    @staticmethod
    def _check_database(connection) -> bool:
        # May throw when sqlar does not exist or it is encrypted without the correct key being specified.
        result = connection.execute("SELECT name FROM sqlar LIMIT 1;").fetchone()
        return result and result[0]

    @staticmethod
    def _find_password(uriPath: str, passwords: List[bytes], salt: bytes):
        for password in passwords:
            # Do the key derivation manually in order to support all characters in passwords, even " and ;.
            # https://stackoverflow.com/a/79657272
            # https://www.zetetic.net/sqlcipher/sqlcipher-api/#PRAGMA_key
            # https://www.zetetic.net/sqlcipher/design/
            keyDerivation = PBKDF2HMAC(
                algorithm=hashes.SHA512(),
                # 256-bit key for 256-bit AES in CBC mode
                length=32,
                # The salt is stored in the first 16 bytes.
                salt=salt,
                # This is the current default. Older versions may have used fewer iterations.
                # It can also be specified with 'PRAGMA kdf_iter'.
                iterations=256_000,
                backend=default_backend(),
            )

            try:
                connection = sqlcipher3.connect(uriPath, uri=True, check_same_thread=False)  # pylint:disable=no-member
                # https://stackoverflow.com/questions/77406316/
                #   how-do-you-safely-pass-values-to-sqlite-pragma-statements-in-python
                # Single quotes after the key= do not work!
                connection.execute(f"""PRAGMA key="x'{keyDerivation.derive(password).hex()}'";""")
                # Unfortunately sqlcipher3 prints lots of output on a wrong password:
                #   2025-06-07 20:42:22.736: ERROR CORE sqlcipher_page_cipher: hmac check failed for pgno=1
                #   2025-06-07 20:42:22.736: ERROR CORE sqlite3Codec: error decrypting page 1 data: 1
                #   2025-06-07 20:42:22.736: ERROR CORE sqlcipher_codec_ctx_set_error 1
                SQLARMountSource._check_database(connection)
                return connection
            except sqlcipher3.DatabaseError:  # pylint:disable=no-member
                pass
        return None

    @staticmethod
    def _convert_to_file_info(rowid: int, mode: int, mtime: int, size: int, linkname: str) -> FileInfo:
        # fmt: off
        return FileInfo(
            size     = max(0, int(size)),
            mtime    = int(mtime),
            mode     = mode,
            linkname = linkname,
            uid      = os.getuid(),
            gid      = os.getgid(),
            userdata = [rowid],
        )
        # fmt: on

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    def _list_names(self, path: str) -> Optional[Iterable[str]]:
        if not self._files:
            return None

        path = path.lstrip('/')
        if path and not path.endswith('/'):
            path = path + '/'

        # Use list instead of set to keep the order.
        names = []
        for name in self._files:
            if not name.startswith(path):
                continue
            name = name[len(path) :].split('/', maxsplit=1)[0]
            if name not in names:
                names.append(name)
        return names

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        if self._files:
            return self._list_names(path)

        pathGlob = "*" if path == "/" else path.strip("/") + "/*"
        return {
            name: SQLARMountSource._convert_to_file_info(rowid, mode, mtime, size, linkname)
            for name, rowid, mode, mtime, size, linkname in self.connection.execute(
                f"SELECT substr(name,(?)),{SQLARMountSource._SQLITE_FILEINFO_COLUMNS} FROM sqlar "
                "WHERE name GLOB (?) AND substr(name,(?)) NOT GLOB '*/*'",
                (len(pathGlob), pathGlob, len(pathGlob)),
            )
        }

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        if self._files:
            return self._list_names(path)

        pathGlob = "*" if path == "/" else path.strip("/") + "/*"
        return dict(
            self.connection.execute(
                "SELECT substr(name,(?)), mode FROM sqlar WHERE name GLOB (?) AND substr(name,(?)) NOT GLOB '*/*'",
                (len(pathGlob), pathGlob, len(pathGlob)),
            )
        )

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        if path == "/":
            return create_root_file_info([-1])

        if self._files:
            path = path.lstrip('/')
            # We assume here that all parent folders to files appear in the table.
            if path not in self._files:
                return None
            rowid = self._files[path]

            result = self.connection.execute(
                f"SELECT {SQLARMountSource._SQLITE_FILEINFO_COLUMNS} FROM sqlar WHERE rowid=(?)", (rowid,)
            ).fetchone()
        else:
            result = self.connection.execute(
                f"SELECT {SQLARMountSource._SQLITE_FILEINFO_COLUMNS} FROM sqlar WHERE name=(?)", (path.strip("/"),)
            ).fetchone()

        return self._convert_to_file_info(*result) if result else None

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        assert fileInfo.userdata
        rowid = fileInfo.userdata[-1]
        assert isinstance(rowid, int)

        result = self.connection.execute("SELECT sz,LENGTH(data) FROM sqlar WHERE rowid == (?);", (rowid,)).fetchone()
        if not result:
            raise FileNotFoundError(f"[SQLARMountSource] rowid: {rowid}")

        # Simply because we can, already try to use the new Python 3.11 sqlite3.Connection.blobopen helper!
        # Does not work for sqlcipher3 0.5.4!
        blob: Any = None
        if sys.version_info >= (3, 11) and hasattr(self.connection, 'blobopen'):
            blob = FixedSqliteBlob(self.connection.blobopen("sqlar", "data", rowid, readonly=True))
        else:
            blob = SQLiteBlobFile(
                self.connection, f"SELECT {{}}data{{}} FROM sqlar WHERE ROWID == {rowid}", size=int(result[1])
            )
        return blob if result[0] == result[1] else rapidgzip.open(blob)

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.connection.close()
        if self._temporaryFilePath:
            self._temporaryFilePath.close()



import io
import os
import stat
from typing import IO, Dict, Iterable, Optional, Union

try:
    import pygit2
except ImportError:
    pygit2 = None  # type: ignore

from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.utils import overrides


class GitMountSource(MountSource):
    """
    Reimplementation from scratch of the very barebones implementation inside fsspec
    because it is slow and "older" versions did not work with pygit2 1.15.

    https://github.com/fsspec/filesystem_spec/blob/master/fsspec/implementations/git.py
    https://github.com/fsspec/filesystem_spec/issues/1708
    """

    enabled = pygit2 is not None

    # pylint: disable=unused-argument
    def __init__(self, path: Optional[str] = None, reference: Optional[str] = None, **kwargs):
        self.repository = pygit2.Repository(path or os.getcwd())
        self.reference = reference or self._get_default_reference(self.repository)
        commit, reference = self.repository.resolve_refish(self.reference)  # type: ignore
        self.tree = commit.tree
        self.commitTime = self.repository[self.repository.head.target].commit_time  # type: ignore
        self.prefix = ""

    @staticmethod
    def _get_default_reference(repository):
        if 'init.defaultBranch' in repository.config:
            return repository.config['init.defaultBranch']

        # Try to find checked out branch.
        for branch in repository.branches:
            if repository.branches[branch].is_head():
                return branch

        for branch in ['master', 'main']:
            if branch in repository.branches:
                return branch

        return 'master'

    def _look_up_path(self, path: str):
        tree = self.tree
        for name in self.prefix.split("/") + path.split("/"):
            if name and isinstance(tree, pygit2.Tree):
                if name not in tree:
                    return None
                tree = tree[name]
        return tree

    @staticmethod
    def _convert_to_file_mode(obj):
        if obj.filemode == pygit2.enums.FileMode.LINK:
            return 0o555 | stat.S_IFLNK
        return 0o555 | (stat.S_IFDIR if isinstance(obj, pygit2.Tree) else stat.S_IFREG)

    def _convert_to_file_info(self, obj, path: str):
        # fmt: off
        return FileInfo(
            size     = obj.size if hasattr(obj, 'size') else 0,
            mtime    = self.commitTime,
            mode     = GitMountSource._convert_to_file_mode(obj),
            linkname = obj.data.decode() if obj.filemode == pygit2.enums.FileMode.LINK else "",
            uid      = os.getuid(),
            gid      = os.getgid(),
            userdata = [path],
        )
        # fmt: on

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    @overrides(MountSource)
    def exists(self, path: str) -> bool:
        return self._look_up_path(path) is not None

    def _list(self, path: str, onlyMode: bool) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        tree = self._look_up_path(path)
        if not isinstance(tree, pygit2.Tree):
            return None
        return {
            obj.name: (
                GitMountSource._convert_to_file_mode(obj)
                if onlyMode
                else self._convert_to_file_info(obj, path + '/' + obj.name)
            )
            for obj in tree
            if obj.name
        }

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        return self._list(path, onlyMode=False)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        return self._list(path, onlyMode=True)

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        obj = self._look_up_path(path)
        return None if obj is None else self._convert_to_file_info(obj, path)

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return 1

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        path = fileInfo.userdata[-1]
        assert isinstance(path, str)
        # TODO Avoid high memory usage for very large files.
        #      Check whether pygit2 even has a kind of streaming API for file contents.
        return io.BytesIO(self._look_up_path(path).data)

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass



import io
import json
import stat
import sys
import tarfile
from timeit import default_timer as timer
from typing import IO, Any, Dict, List, Optional, Tuple, Union

from ratarmountcore.formats import FileFormatID, replace_format_check
from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource
from ratarmountcore.SQLiteIndex import SQLiteIndex
from ratarmountcore.utils import overrides

try:
    import py7zr

    if not py7zr.__version__.startswith("1.0"):
        raise ImportError("Wrong version found!")

    from py7zr.io import Py7zIO, WriterFactory

    class CachedFile(Py7zIO):
        def __init__(self, offset: int = 0, maxSize: int = sys.maxsize):
            self.file = io.BytesIO()
            self._offset = offset
            self._maxSize = maxSize
            self._position = 0

        @overrides(Py7zIO)
        def write(self, s: Union[bytes, bytearray]) -> int:
            toSkip = self._offset - self._position
            self._position += len(s)
            if toSkip < len(s) and self.file.tell() < self._maxSize:
                toRead = self._maxSize - self.file.tell()
                self.file.write(s[toSkip : toSkip + toRead])
            return self._position

        @overrides(Py7zIO)
        def read(self, size: Optional[int] = None) -> bytes:
            raise NotImplementedError

        @overrides(Py7zIO)
        def flush(self) -> None:
            pass

        @overrides(Py7zIO)
        def seek(self, offset: int, whence: int = 0) -> int:  # pylint: disable=unused-argument
            return offset

        @overrides(Py7zIO)
        def size(self) -> int:
            return self.file.tell()

    class CachedFileFactory(WriterFactory):
        # This factory only seems so useless because we are only extracting single files.
        # It makes sense when extracting multiple files.
        def __init__(self, offset: int = 0, maxSize: int = sys.maxsize):
            self._offset = offset
            self._maxSize = maxSize
            self.products: Dict[str, CachedFile] = {}

        def create(self, filename) -> Py7zIO:
            product = CachedFile(self._offset, self._maxSize)
            self.products[filename] = product
            return product

    def open_in_memory(archive, target: str, offset: int = 0, maxSize: int = sys.maxsize) -> IO[bytes]:
        factory = CachedFileFactory(offset=offset, maxSize=maxSize)
        archive.reset()
        archive.extract(targets=[target], factory=factory)
        file = factory.products[target].file
        file.seek(0)
        return file

    # https://github.com/miurahr/py7zr/issues/659#issuecomment-2954260661
    replace_format_check(FileFormatID.SEVEN_ZIP, py7zr.is_7zfile)  # type: ignore

except ImportError:
    py7zr = None  # type: ignore


class Py7zrMountSource(SQLiteIndexMountSource):
    # fmt: off
    def __init__(
        self,
        fileOrPath             : Union[str, IO[bytes]],
        writeIndex             : bool                      = False,
        clearIndexCache        : bool                      = False,
        indexFilePath          : Optional[str]             = None,
        indexFolders           : Optional[List[str]]       = None,
        encoding               : str                       = tarfile.ENCODING,
        verifyModificationTime : bool                      = False,
        printDebug             : int                       = 0,
        indexMinimumFileCount  : int                       = 1000,
        **options
    ) -> None:
        self.archiveFilePath        = fileOrPath if isinstance(fileOrPath, str) else None
        self.encoding               = encoding
        self.verifyModificationTime = verifyModificationTime
        self.printDebug             = printDebug
        self.options                = options
        # fmt: on

        # TODO For now, 'transform' is not supported because we need the exact path to open the file and there
        #      currently is no SQLite table column to store this information in.
        # TODO I doubt that symbolic links work because py7zr.FileInfo does not have information regarding links.

        def open_file(password: Optional[Union[str, bytes]] = None):
            return py7zr.SevenZipFile(
                # https://github.com/miurahr/py7zr/issues/659#issuecomment-2954260661
                fileOrPath,  # type: ignore
                password=None if password is None else (password if isinstance(password, str) else password.decode()),
            )

        self.fileObject = Py7zrMountSource._find_password(open_file, options.get("passwords", []))

        # Force indexes in memory because:
        #  - I have no idea what ID to write into offset or offsetheader. The "ID" for the py7zr interface
        #    is the "filename" (path). Storing a string in the int 'offset' column is not a got idea.
        #  - The py7zr interface seems to lack a way to query information about symbolic links.
        indexFilePath = ':memory:'
        super().__init__(
            SQLiteIndex(
                indexFilePath,
                indexFolders=indexFolders,
                archiveFilePath=self.archiveFilePath,
                encoding=self.encoding,
                printDebug=self.printDebug,
                indexMinimumFileCount=indexMinimumFileCount,
                backendName='Py7zrMountSource',
            ),
            clearIndexCache=clearIndexCache,
            checkMetadata=self._check_metadata,
        )

        isFileObject = not isinstance(fileOrPath, str)

        if self.index.index_is_loaded():
            self.index.reload_index_read_only()
        else:
            # Open new database when we didn't find an existing one.
            # Simply open in memory without an error even if writeIndex is True but when not indication
            # for a index file location has been given.
            if writeIndex and (indexFilePath or not isFileObject):
                self.index.open_writable()
            else:
                self.index.open_in_memory()

            self._create_index()
            if self.index.index_is_loaded():
                self._store_metadata()
                self.index.reload_index_read_only()

    def _store_metadata(self) -> None:
        argumentsToSave = ['encoding']
        argumentsMetadata = json.dumps({argument: getattr(self, argument) for argument in argumentsToSave})
        self.index.store_metadata(argumentsMetadata, self.archiveFilePath)

    def _convert_to_row(self, info) -> Tuple:
        mode = 0o777 | (stat.S_IFDIR if info.is_directory else stat.S_IFREG)
        mtime = info.creationtime.timestamp()
        path, name = SQLiteIndex.normpath(info.filename).rsplit("/", 1)

        # fmt: off
        fileInfo : Tuple = (
            path              ,  # 0  : path
            name              ,  # 1  : file name
            0                 ,  # 2  : header offset
            0                 ,  # 3  : data offset
            info.uncompressed ,  # 4  : file size
            mtime             ,  # 5  : modification time
            mode              ,  # 6  : file mode / permissions
            0                 ,  # 7  : TAR file type. Currently unused. Overlaps with mode
            # Abuse the linkname to store the "ID" for file access.
            info.filename     ,  # 8  : linkname
            0                 ,  # 9  : user ID
            0                 ,  # 10 : group ID
            False             ,  # 11 : is TAR (unused?)
            False             ,  # 12 : is sparse
            False             ,  # 13 : is generated (parent folder)
            0                 ,  # 14 : recursion depth
        )
        # fmt: on

        return fileInfo

    def _create_index(self) -> None:
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} ...")
        t0 = timer()

        self.index.ensure_intermediary_tables()
        self.index.set_file_infos([self._convert_to_row(info) for info in self.fileObject.list()])

        # Resort by (path,name). This one-time resort is faster than resorting on each INSERT (cache spill)
        if self.printDebug >= 2:
            print("Resorting files by path ...")

        self.index.finalize()

        t1 = timer()
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} took {t1 - t0:.2f}s")

    @staticmethod
    def _find_password(openFile, passwords):
        fileobj = openFile()

        # If headers are encrypted, then infolist will simply return an empty list!
        files = fileobj.list()
        if not files:
            for password in passwords:
                fileobj = openFile(password)
                files = fileobj.list()
                if files:
                    return fileobj

        # If headers are not encrypted, then try out passwords by trying to open the first file.
        files = [file for file in files if not file.is_directory and file.uncompressed > 0]
        if not files:
            return fileobj

        for password in [None, *passwords]:
            fileobj = openFile(password)
            try:
                open_in_memory(fileobj, files[0].filename)
                return fileobj
            except Exception:
                pass

        raise RuntimeError("Could not find a matching password!")

    @overrides(SQLiteIndexMountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        super().__exit__(exception_type, exception_value, exception_traceback)
        self.fileObject.close()

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        return open_in_memory(self.fileObject, fileInfo.linkname)

    def _check_metadata(self, metadata: Dict[str, Any]) -> None:
        """Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible."""
        SQLiteIndex.check_archive_stats(self.archiveFilePath, metadata, self.verifyModificationTime)

        if 'arguments' in metadata:
            SQLiteIndex.check_metadata_arguments(json.loads(metadata['arguments']), self, argumentsToCheck=['encoding'])



# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import io
import os
import sqlite3
import sys
import tempfile

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.SQLiteBlobFile import SQLiteBlobFile, SQLiteBlobsFile  # noqa: E402

testData = b"1234567890"
tmpFile = tempfile.TemporaryFile()
tmpFile.write(testData)


randomTestData = os.urandom(128 * 1024)
randomTmpFile = tempfile.TemporaryFile()
randomTmpFile.write(randomTestData)


class TestLambdaReaderFile:
    @staticmethod
    def test_single_blob():
        db = sqlite3.connect(":memory:")
        db.execute('CREATE TABLE gzipindexes ( data BLOB )')
        db.execute('INSERT INTO gzipindexes VALUES (?)', (randomTestData,))

        expression = "SELECT {}data{} FROM gzipindexes WHERE ROWID == 1"

        assert SQLiteBlobFile.read_blob_part(db, expression, 0, 10) == randomTestData[0:10]
        assert SQLiteBlobFile.read_blob_part(db, expression, 300, 128) == randomTestData[300 : 300 + 128]
        assert SQLiteBlobFile.read_blob_part(db, expression, 1000, 1000) == randomTestData[1000:2000]
        assert SQLiteBlobFile.read_blob_part(db, expression, 0, len(randomTestData)) == randomTestData

        file = SQLiteBlobFile(db, expression)
        assert file.tell() == 0

        assert file.read(10) == randomTestData[0:10]
        assert file.tell() == 10

        assert file.seek(300) == 300
        assert file.tell() == 300
        assert file.read(128) == randomTestData[300 : 300 + 128]
        assert file.tell() == 300 + 128

        assert file.seek(1000) == 1000
        assert file.tell() == 1000
        assert file.read(1000) == randomTestData[1000:2000]
        assert file.tell() == 2000

        assert file.seek(0) == 0
        assert file.read() == randomTestData
        assert file.read() == b""

    @staticmethod
    def test_two_blobs():
        db = sqlite3.connect(":memory:")
        db.execute('CREATE TABLE gzipindexes ( data BLOB )')
        db.executemany('INSERT INTO gzipindexes VALUES (?)', ((b"bar",), (b"foo",)))

        file = SQLiteBlobsFile(db, 'gzipindexes', 'data')

        assert file.read(1) == b"b"
        assert file.tell() == 1

        assert file.seek(3) == 3
        assert file.tell() == 3

        assert file.read(2) == b"fo"
        assert file.tell() == 5

        assert file.seek(-4, io.SEEK_END) == 2
        assert file.tell() == 2

        assert file.read() == b"rfoo"
        assert file.tell() == 6

        assert file.read() == b""
        assert file.tell() == 6



import dataclasses
import os
import stat
import time
from abc import ABC, abstractmethod
from typing import IO, Any, Dict, Iterable, List, Optional, Union


@dataclasses.dataclass
class FileInfo:
    # fmt: off
    size     : int
    mtime    : float
    mode     : int
    linkname : str
    uid      : int
    gid      : int
    # By convention this is a list and MountSources should only read the last element and before forwarding the
    # FileInfo to a possibly recursively "mounted" MountSource, remove that last element belonging to it.
    # This way an arbitrary amount of userdata can be stored and it should be decidable which belongs to whom in
    # a chain of MountSource objects.
    userdata : List[Any]
    # fmt: on

    def clone(self):
        copied = dataclasses.replace(self)
        # Make a new userdata list but do not do a full deep copy because some MountSources put references
        # to MountSources into userdata and those should and can not be deep copied.
        copied.userdata = self.userdata[:]
        return copied


class MountSource(ABC):
    """
    Generic class representing a mount point. It's basically like the FUSE API but boiled down
    to the necessary methods for ratarmount.

    Similar, to FUSE, all paths should have a leading '/'.
    If there is is no leading slash, behave as if there was one.
    """

    @abstractmethod
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        pass

    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        """
        This function can and should be overwritten with something that is faster than list
        because only a simple path -> mode mapping needs to be returned, not all file metadata.
        This method is custom-tailored for FUSE readdir, i.e., the returned mode is not guaranteed
        to include file permissions, only the S_IFREG, S_IFLINK, S_IFDIR flags, maybe all S_IFMT flags
        in the future.
        """
        result = self.list(path)
        if isinstance(result, dict):
            return {path: fileInfo.mode for path, fileInfo in result.items()}
        return result

    @abstractmethod
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        pass

    @abstractmethod
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        """
        buffering : Behaves similarly to Python's built-in open call. A value of 0 should disable buffering.
                    Any value larger than 1 should be the buffer size. The default of -1 may result in
                    a default buffer size equal to the file(system)'s block size or Python's io.DEFAULT_BUFFER_SIZE.
        """

    def statfs(self) -> Dict[str, Any]:
        """
        Returns a dictionary with keys named like the POSIX statvfs struct.
        https://pubs.opengroup.org/onlinepubs/009695399/basedefs/sys/statvfs.h.html
        Keys may be missing. Either an empty dictionary should be returned, or at least f_bsize and f_namemax
        should be initialized because that's what libfuse returns per default if the statfs call is not implemented.
        If statfs is not implemented / the returned dictionary call is empty, libfuse will return default values:
            {'f_bsize': 512, 'f_namemax': 255}
        https://github.com/libfuse/libfuse/blob/373ddc7eae7b0c684fc4ab29d8addfa3b9e99e1e/lib/fuse.c#L1962-L1975
        """
        return {}

    def versions(self, path: str) -> int:
        return 1 if self.exists(path) else 0

    def read(self, fileInfo: FileInfo, size: int, offset: int) -> bytes:
        # Because we only do a single seek before closing the file again, buffering makes no sense.
        with self.open(fileInfo, buffering=0) as file:
            file.seek(offset)
            return file.read(size)

    @abstractmethod
    def is_immutable(self) -> bool:
        """
        Should return True if the mount source is known to not change over time in order to allow for optimizations.
        Meaning, all interface methods should return the same results given the same arguments at any time.
        """

    def get_mount_source(self, fileInfo: FileInfo):
        """
        Returns the direct mount source to which the fileInfo belongs, a mount source specific file info,
        and the mount point of the returned mount source in respect to this (self) MountSource.
        """
        return '/', self, fileInfo

    def exists(self, path: str) -> bool:
        return self.lookup(path) is not None

    def is_dir(self, path: str):
        fileInfo = self.lookup(path)
        return fileInfo is not None and stat.S_ISDIR(fileInfo.mode)

    # pylint: disable=unused-argument
    def list_xattr(self, fileInfo: FileInfo) -> List[str]:
        """
        Should return list of extended file attribute keys for the given fileInfo.
        """
        return []

    # pylint: disable=unused-argument
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        """
        Should return value for given fileInfo and extended file attribute key.
        """
        return None

    def __enter__(self):
        return self

    # If the derived MountSource opens some file object or similar in its constructor
    # then it should override this and close the file object.
    @abstractmethod
    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass


def create_root_file_info(userdata: List[Any]):
    # fmt: off
    return FileInfo(
        size     = 0,
        mtime    = int(time.time()),
        mode     = 0o777 | stat.S_IFDIR,
        linkname = "",
        uid      = os.getuid(),
        gid      = os.getgid(),
        userdata = userdata,
    )
    # fmt: on


def merge_statfs(values: Iterable[Dict[str, Any]], printDebug: int = 0):
    result = {}
    for statfs in values:
        for key, value in statfs.items():
            if key not in result:
                result[key] = value
                continue

            if key in ('f_bsize', 'f_frsize'):
                result[key] = max(result[key], value)
                continue

            if key == 'f_namemax':
                result[key] = min(result[key], value)
                continue

            if result[key] != value and printDebug >= 1:
                print(f"[Warning] Failed to merge statfs values ({value}, {result[key]}) for key: {key}.")
    return result



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import stat
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.formats.rar import RarMountSource  # noqa: E402


def find_test_file(relativePathOrName):
    for i in range(3):
        path = os.path.sep.join([".."] * i + ["tests", relativePathOrName])
        if os.path.exists(path):
            return path
    return relativePathOrName


class TestRarMountSource:
    @staticmethod
    def test_simple_usage():
        with RarMountSource(find_test_file('folder-symlink.rar')) as mountSource:
            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)
                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'iriya\n'

            # Links are not resolved by the mount source but by FUSE, i.e., descending into a link to a folder
            # will not work. This behavior may change in the future.
            for linkPath in ['/foo/jet']:
                assert mountSource.lookup(linkPath)
                assert mountSource.versions(linkPath) == 1
                assert not mountSource.list(linkPath)
                with mountSource.open(mountSource.lookup(linkPath)) as file:
                    # Contents of symlink is the symlink destination itself.
                    assert file.read() == b'fighter'

    # TODO 'transform' does not work. Could be made to work easily when refactoring it to use SQLiteIndex



import bisect
import io
from typing import IO, Callable, List, Optional, Tuple

from .utils import FixedRawIOBase, _DummyContext, overrides


class RawStenciledFile(FixedRawIOBase):
    # For a reference implementation based on RawIOBase, see "class SocketIO(io.RawIOBase)" in:
    #   https://github.com/python/cpython/blob/00ffdf27367fb9aef247644a96f1a9ffb5be1efe/Lib/socket.py#L683
    # or others implementations inside cpython:
    #   https://github.com/python/cpython/blob/00ffdf27367fb9aef247644a96f1a9ffb5be1efe/Lib/_compression.py#L33
    # For the internals of RawIOBase and others, see:
    #   https://github.com/python/cpython/tree/main/Lib/_pyio.py#L619
    """A file abstraction layer giving a stenciled view to an underlying file."""

    def __init__(
        self,
        fileStencils: List[Tuple[IO, int, int]],
        fileObjectLock=None,
    ) -> None:
        """
        stencils: A list tuples specifying the offset and length of the underlying file to use.
                  The order of these tuples will be kept.
                  The offset must be non-negative and the size must be positive.
        fileobj: (deprecated) Only either fileobj and stencils or fileStencils may be specified
        stencils: (deprecated) Only either fileobj and stencils or fileStencils may be specified
        fileStencils: Contains a list of (file object, offset, size) tuples. The offset and size
                      can be seen as a cut-out of the file object. All cut-outs are joined
                      together in the order of the list. Note that multiple cut-outs into the
                      same file object may be given by simply specifying the file object multiple
                      times in the list.

        Examples:
            stencil = [(5,7)]
                Makes a new 7B sized virtual file starting at offset 5 of fileobj.
            stencil = [(0,3),(5,3)]
                Make a new 6B sized virtual file containing bytes [0,1,2,5,6,7] of fileobj.
            stencil = [(0,3),(0,3)]
                Make a 6B size file containing the first 3B of fileobj twice concatenated together.
        """
        super().__init__()

        self.offset = 0
        self.fileObjectLock = fileObjectLock
        self.offsets: List[int] = []
        self.sizes: List[int] = []
        self.fileObjects: List[IO] = []

        if fileStencils:
            self.fileObjects, self.offsets, self.sizes = zip(*fileStencils)

        # Check whether values make sense
        for offset in self.offsets:
            assert offset >= 0
        for size in self.sizes:
            assert size >= 0

        # Filter out zero-sized regions (or else we would have to skip them inside 'readinto' in order to not
        # return an empty reply even though the end of file has not been reached yet!)
        selectedStencils = [i for i, size in enumerate(self.sizes) if size > 0]
        self.offsets = [self.offsets[i] for i in selectedStencils]
        self.sizes = [self.sizes[i] for i in selectedStencils]
        self.fileObjects = [self.fileObjects[i] for i in selectedStencils]

        # Check for readability
        for fileObject in self.fileObjects:
            if not fileObject.readable():
                raise ValueError("All file objects to be joined must be readable")

        # Calculate cumulative sizes
        self.cumsizes = [0]
        for size in self.sizes:
            self.cumsizes.append(self.cumsizes[-1] + size)

        # Seek to the first stencil offset in the underlying file so that "read" will work out-of-the-box
        self.seek(0)

    def _find_stencil(self, offset: int) -> int:
        """
        Return index to stencil where offset belongs to. E.g., for stencils [(3,5),(8,2)], offsets 0 to
        and including 4 will still be inside stencil (3,5), i.e., index 0 will be returned. For offset 6,
        index 1 would be returned because it now is in the second contiguous region / stencil.
        """
        # bisect_left( value ) gives an index for a lower range: value < x for all x in list[0:i]
        # Because value >= 0 and list starts with 0 we can therefore be sure that the returned i>0
        # Consider the stencils [(11,2),(22,2),(33,2)] -> cumsizes [0,2,4,6]. Seek to offset 2 should seek to 22.
        assert offset >= 0
        i = bisect.bisect_left(self.cumsizes, offset + 1) - 1
        assert i >= 0
        return i

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass

    @overrides(io.RawIOBase)
    def seekable(self) -> bool:
        return all(fobj.seekable() for fobj in self.fileObjects)

    @overrides(io.RawIOBase)
    def readable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def readinto(self, buffer):
        """Generic implementation which uses read."""
        with memoryview(buffer) as view, view.cast("B") as byteView:  # type: ignore
            readBytes = self.read(len(byteView))
            byteView[: len(readBytes)] = readBytes
        return len(readBytes)

    def _read1_unlocked(self, size):
        assert size >= 0

        i = self._find_stencil(self.offset)
        if i >= len(self.sizes):
            return b''

        # Note that seek and read of the file object itself do not seem to check against this and
        # instead lead to a segmentation fault in the multithreading tests.
        if self.fileObjects[i].closed:
            raise ValueError("A closed file cannot be read from!")

        offsetInsideStencil = self.offset - self.cumsizes[i]
        assert offsetInsideStencil >= 0
        assert offsetInsideStencil < self.sizes[i]
        self.fileObjects[i].seek(self.offsets[i] + offsetInsideStencil, io.SEEK_SET)

        # Read as much as requested or as much as the current contiguous region / stencil still contains.
        readableSize = min(size, self.sizes[i] - (self.offset - self.cumsizes[i]))
        tmp = self.fileObjects[i].read(readableSize)
        self.offset += len(tmp)
        return tmp

    @overrides(io.RawIOBase)
    def read(self, size: int = -1) -> bytes:
        if size == -1:
            size = self.cumsizes[-1] - self.offset
        elif size == 0:
            return b''

        result = b''
        with self.fileObjectLock or _DummyContext():
            while len(result) < size:
                tmp = self._read1_unlocked(size - len(result))
                if not tmp:
                    break
                result += tmp

        return result

    @overrides(io.RawIOBase)
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if whence == io.SEEK_CUR:
            self.offset += offset
        elif whence == io.SEEK_END:
            self.offset = self.cumsizes[-1] + offset
        elif whence == io.SEEK_SET:
            self.offset = offset

        if self.offset < 0:
            raise ValueError("Trying to seek before the start of the file!")
        if self.offset >= self.cumsizes[-1]:
            return self.offset

        return self.offset

    @overrides(io.RawIOBase)
    def tell(self) -> int:
        return self.offset


class RawJoinedFileFromFactory(io.RawIOBase):
    def __init__(self, file_object_factories: List[Callable[[], IO[bytes]]], file_lock=None) -> None:
        """
        Similar to StenciledFile but instead of joining a list of file objects, which neccessitates keeping all
        files open, this class opens each file on demand and only keeps one file open. This is useful to avoid
        exceeding the system limit for the number of opened I/O handles.
        """
        io.RawIOBase.__init__(self)

        self.offset = 0
        self.fileObjectLock = file_lock
        # Stores index and file object of currently opened file object. It basically is a cache for the factories.
        self.fileObject: Optional[Tuple[int, IO[bytes]]] = None

        self.sizes: List[int] = []
        self.factories: List[Callable[[], IO[bytes]]] = []
        self.cumsizes = [0]
        self._seekable = True
        for factory in file_object_factories:
            with factory() as file:
                if not file.seekable():
                    self._seekable = False
                if not file.readable():
                    raise ValueError("All file objects to be joined must be readable")

                size = file.seek(0, io.SEEK_END)
                if size is None or size < 0:
                    raise ValueError(f"Failed to query size for factory: {factory}")

                # Filter out zero-sized regions (or else we would have to skip them inside 'readinto' in order to not
                # return an empty reply even though the end of file has not been reached yet!)
                if size > 0:
                    self.sizes.append(size)
                    self.factories.append(factory)
                    self.cumsizes.append(self.cumsizes[-1] + size)

    def _find_stencil(self, offset: int) -> int:
        # See StenciledFile._find_stencil
        assert offset >= 0
        i = bisect.bisect_left(self.cumsizes, offset + 1) - 1
        # i might be 0 when self.cumsizes is empty but even for no fileStencils, it is initialized with [0]
        assert i >= 0
        return i

    def _get_file_object(self, index: int) -> IO[bytes]:
        if self.fileObject and self.fileObject[0] == index:
            return self.fileObject[1]

        self._close_fileobj()
        self.fileObject = index, self.factories[index]()
        return self.fileObject[1]

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    def _close_fileobj(self) -> None:
        if self.fileObject and not self.fileObject[1].closed:
            self.fileObject[1].close()

    @overrides(io.RawIOBase)
    def close(self) -> None:
        self._close_fileobj()
        super().close()

    @overrides(io.RawIOBase)
    def seekable(self) -> bool:
        return self._seekable

    @overrides(io.RawIOBase)
    def readable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def readinto(self, buffer):
        """Generic implementation which uses read."""
        with memoryview(buffer) as view, view.cast("B") as byteView:  # type: ignore
            readBytes = self.read(len(byteView))
            byteView[: len(readBytes)] = readBytes
        return len(readBytes)

    @overrides(io.RawIOBase)
    def read(self, size: int = -1) -> bytes:
        # See StenciledFile.read
        if size == -1:
            size = self.cumsizes[-1] - self.offset

        # This loop works in a kind of leapfrog fashion. On each even loop iteration it seeks to the next stencil
        # and on each odd iteration it reads the data and increments the offset inside the stencil!
        result = b''
        i = self._find_stencil(self.offset)
        if i >= len(self.sizes):
            return result

        with self.fileObjectLock or _DummyContext():
            fileObject = self._get_file_object(i)

            # Note that seek and read of the file object itself do not seem to check against this and
            # instead lead to a segmentation fault in the multithreading tests.
            if fileObject.closed:
                raise ValueError("A closed file can't be read from!")

            offsetInsideStencil = self.offset - self.cumsizes[i]
            assert offsetInsideStencil >= 0
            assert offsetInsideStencil < self.sizes[i]
            fileObject.seek(offsetInsideStencil, io.SEEK_SET)

            # Read as much as requested or as much as the current contiguous region / stencil still contains
            readableSize = min(size, self.sizes[i] - (self.offset - self.cumsizes[i]))
            tmp = fileObject.read(readableSize)
            self.offset += len(tmp)
            result += tmp

        return result

    @overrides(io.RawIOBase)
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        """Generic implementation that simply updates self.offset nothing else."""
        if whence == io.SEEK_CUR:
            self.offset += offset
        elif whence == io.SEEK_END:
            self.offset = self.cumsizes[-1] + offset
        elif whence == io.SEEK_SET:
            self.offset = offset

        # Same behavior for seeking before the file begin and after the file end as io.BytesIO.
        self.offset = max(self.offset, 0)
        if self.offset >= self.cumsizes[-1]:
            return self.offset

        return self.offset

    @overrides(io.RawIOBase)
    def tell(self) -> int:
        return self.offset


class StenciledFile(io.BufferedReader):
    def __init__(
        self, fileStencils: List[Tuple[IO, int, int]], fileObjectLock=None, bufferSize=io.DEFAULT_BUFFER_SIZE
    ) -> None:
        """
        bufferSize: Gets forwarded to io.BufferedReader.__init__ buffer_size argument and has the same semantic,
                    i.e., must be > 0. If it should be unbuffered, use RawStenciledFile directly instead.
        """
        super().__init__(RawStenciledFile(fileStencils, fileObjectLock), buffer_size=bufferSize)


class JoinedFile(io.BufferedReader):
    def __init__(self, file_objects: List[IO], file_lock=None, buffer_size: int = io.DEFAULT_BUFFER_SIZE) -> None:
        sizes = [fobj.seek(0, io.SEEK_END) for fobj in file_objects]
        for fobj, size in zip(file_objects, sizes):
            if size is None:
                raise ValueError("Failed to query size of file object:", fobj)

        fileStencils = [(fobj, 0, size or 0) for fobj, size in zip(file_objects, sizes)]
        super().__init__(RawStenciledFile(fileStencils=fileStencils, fileObjectLock=file_lock), buffer_size=buffer_size)


class JoinedFileFromFactory(io.BufferedReader):
    def __init__(
        self, file_object_factories: List[Callable[[], IO]], file_lock=None, buffer_size: int = io.DEFAULT_BUFFER_SIZE
    ) -> None:
        """
        Similar to JoinedFile but instead of joining a list of file objects, which neccessitates keeping all
        files open, this class opens each file on demand and only keeps one file open. This is useful to avoid
        exceeding the system limit for the number of opened I/O handles.
        """
        super().__init__(RawJoinedFileFromFactory(file_object_factories, file_lock), buffer_size=buffer_size)


class LambdaReaderFile(io.RawIOBase):
    """Creates a file abstraction from a single read(offset, size) function."""

    def __init__(self, rawRead: Callable[[int, int], bytes], size: int) -> None:
        """rawRead: Function which returns bytes for (offset, size) input."""
        io.RawIOBase.__init__(self)

        self.offset = 0
        self.rawRead = rawRead
        self.size = size

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass

    @overrides(io.RawIOBase)
    def seekable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def readable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def readinto(self, buffer):
        with memoryview(buffer) as view, view.cast("B") as byteView:  # type: ignore
            readBytes = self.read(len(byteView))
            byteView[: len(readBytes)] = readBytes
        return len(readBytes)

    @overrides(io.RawIOBase)
    def read(self, size: int = -1) -> bytes:
        result = self.rawRead(self.offset, self.size if size == -1 else size)
        self.offset += len(result)
        return result

    @overrides(io.RawIOBase)
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if whence == io.SEEK_CUR:
            self.offset += offset
        elif whence == io.SEEK_END:
            self.offset = self.size + offset
        elif whence == io.SEEK_SET:
            self.offset = offset

        if self.offset < 0:
            raise ValueError("Trying to seek before the start of the file!")
        if self.offset >= self.size:
            return self.offset

        return self.offset

    @overrides(io.RawIOBase)
    def tell(self) -> int:
        return self.offset
